// SPDX-FileCopyrightText: 2010-2022 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef REGISTRATIONPAGE_H
#define REGISTRATIONPAGE_H

// Local includes
#include "shared/RegistrationColumns.h"
#include "shared/Draw.h"

// Qt includes
#include <QWidget>
#include <QIcon>

// Local classes
class SharedObjects;
class StringTools;
class Database;
class TournamentSettings;
class SearchEngine;
class RegistrationDelegate;
class TitleMenu;
class SearchWidget;
class MarkersWidget;
class MarkersPopup;
class Settings;
class Server;
class Client;
class DrawPopup;
class DrawOverviewPage;
class PhoneticSearchButton;
class DismissableMessage;

// Qt classes
class QGroupBox;
class QLabel;
class QLineEdit;
class QPushButton;
class QTableWidget;
class QAction;
class QMenu;
class QTableWidgetItem;
class QToolButton;
class QMessageBox;

namespace rc = RegistrationColumns;

class RegistrationPage : public QWidget
{
    Q_OBJECT

public:
    explicit RegistrationPage(QWidget *parent, SharedObjects *sharedObjects);
    void reload(bool updateMarkersWidget = true);
    void reset();
    bool absentsPresent() const;
    bool singlesPresent() const;
    void setDrawOverviewPage(DrawOverviewPage *drawOverviewPage);
    void checkUpdateDrawOverviewPage(bool ownChange, const QString &target = QString(),
                                     bool deleted = false) const;
    void setColumnsVisible(const QVector<rc::Column> &columns);
    void requestSetColumnVisible(rc::Column column, bool state);
    bool registerPhoneticSearchEnabled() const;
    bool drawAdjustmentNeededState() const;

    void serverStarted(Server *server);
    void serverStopped();
    void clientConnected(Client *client);
    void clientDisconnected();
    void updateChecksum();
    MarkersWidget *markersWidget() const;
    void processNetworkChange(int clientId = 0);
    void setProcessingRequest(bool processing, const QString &scrollToTarget = QString());
    void networkFinishRegistration(int clientId = 0);
    void emitDrawAdjustmentNeeded() const;
    void askShowDrawColumn();

    const QString &checksum() const;
    const QVector<QString> &allRegistrations() const;
    const QVector<Draw::Seat> &allDraws() const;
    const QVector<int> &incompleteTables() const;
    const QVector<int> &tableGaps() const;

public slots:
    void updatePlayersList();
    void scoreChanged();
    void finishRegistration();
    void resetSearch();
    void adoptDrawAdjustment(const QVector<QString> &names, const QVector<Draw::Seat> &seats);
    void requestRejected();
    void requestSetHeaderVisible(bool visible);

signals:
    void statusUpdate(const QString &message);
    void playersCountChanged(int count);
    void registrationsAvailable(bool state);
    void nameEdited();
    void networkTournamentStarted(bool state);
    void drawChanged();
    void showDrawOverviewPage();
    void showAdjustDrawDialog();
    void drawAdjustmentNeeded(bool state) const;
    void listWasModified();
    void disqualificationChanged();
    void columnVisibilityChanged(rc::Column column, bool state);
    void headerVisibilityChanged(bool state);

private slots:
    void registerPlayers();
    void editEntered(const QString &originalName, const QString &name);
    void showContextMenu(QPoint point);
    void scrollTo(const QString &text);
    void deletePlayers();
    void updateSearch(const QString &searchText);
    void markersChanged();
    void updateMarker(const QString &markerName, int markerId);
    void listUpdateMarker(const QString &markerName, int markerId);
    void listRemoveMarker(const QString &markerName, int markerId);
    void deleteMarked(const QString &markerName, int markerId);
    void deleteUnmarked();
    void deleteAll();
    void updateVisibility();
    void markersVisibilityChanged();
    void setColumnVisible(rc::Column column);
    void checkShowPopup(int row, int column);
    void highlightItem(QTableWidgetItem *item);
    void resetHighlight();
    void assignPlayer(const QString &name);
    void updateDraw(const Draw::Seat &seat);
    void drawAllSeats();
    void deleteAllDraws();
    void processStatusLinkClick(const QString &anchor);
    void showDrawSettings();
    void adoptSearchText();
    void showHeaderContextMenu(const QPoint &point);
    void setHideHeader(bool state);
    void disqualifyPlayers();
    void unDisqualifyPlayers();

private: // Functions
    void listModified();
    bool isRowVisible(int row) const;
    void updateStatusLabel();
    bool confirmListDeleteAction(const QString &title, const QString &text);
    void checkDisplayBackup(const QString &title);
    void cacheChecksum();
    bool checkCachedChecksum(const QString &title);
    void hideContextMenus();
    void displayMarkedEntry();
    bool checkForExactMatch(const QString &name);
    void setListDraw(const QVector<Draw::Seat> &draw);
    QString takenTablesString(int tables, bool definitive) const;
    int visibleRow(const QTableWidgetItem *item) const;
    void itemError(const char *funcInfoText, const QString &text);
    void setDisqualified(int round);

private: // Variables
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    SearchEngine *m_searchEngine;
    DismissableMessage *m_dismissableMessage;

    QGroupBox *m_newPlayersBox;
    QLabel *m_nameLabel;
    QLineEdit *m_playersName;
    PhoneticSearchButton *m_registerPhoneticSearch;
    QPushButton *m_registerButton;
    QTimer *m_resetHighlightTimer;
    QGroupBox *m_playersBox;
    QTableWidget *m_playersList;
    QIcon m_markerIcon;
    QIcon m_drawIcon;
    QLabel *m_statusLabel;
    QLabel *m_isSearchLabel;
    QHash<rc::Column, QAction *> m_showColumn;
    RegistrationDelegate *m_delegate;
    SearchWidget *m_searchWidget;
    QToolButton *m_adoptSearchText;
    MarkersWidget *m_markersWidget;
    MarkersPopup *m_markersPopup;
    DrawPopup *m_drawPopup;
    DrawOverviewPage *m_drawOverviewPage = nullptr;
    QMessageBox *m_fuzzyMatchWarning = nullptr;

    QVector<QString> m_allRegistrations;
    QVector<int> m_allSetMarkers;
    QVector<Draw::Seat> m_allDraws;
    QVector<int> m_allDisqualifications;

    QTableWidgetItem *m_highlightedItem;
    int m_editedRow;
    QString m_selectedName;
    QVector<int> m_incompleteTables;
    QVector<int> m_tableGaps;
    QBrush m_defaultForeground;
    QBrush m_highlightedForeground;
    QBrush m_highlightedBackground;
    QBrush m_lastForeground;
    QBrush m_lastBackground;
    bool m_drawAdjustmentNeeded = false;

    bool m_showHideHeaderHint = false;

    // Context menus

    QPushButton *m_menuButton;
    TitleMenu *m_columnsMenu;
    QAction *m_hideHeaderAction;

    TitleMenu *m_contextMenu;

    QMenu *m_assignMenu;

    QAction *m_editAction;
    QAction *m_deleteAction;

    QAction *m_setOnlyMarkerAction;
    QMenu *m_setMarkerMenu;
    QAction *m_removeMarkerAction;

    QAction *m_setDrawAction;
    QAction *m_deleteDrawAction;

    QMenu *m_disqualificationMenu;
    QAction *m_disqualifyAction;
    QAction *m_unDisqualifyAction;

    TitleMenu *m_listMenu;

    QAction *m_listSetOnlyMarkerAction;
    QMenu *m_listSetMarkerMenu;
    QAction *m_listRemoveOnlyMarkerAction;
    QMenu *m_listRemoveMarkerMenu;
    QAction *m_listRemoveAllMarkersAction;

    QAction *m_deleteOnlyMarkedAction;
    QMenu *m_deleteMarkedMenu;
    QAction *m_deleteUnmarkedAction;
    QAction *m_deleteAllAction;

    QAction *m_deleteAllDrawsAction;

    // Network
    Server *m_server = nullptr;
    Client *m_client = nullptr;
    bool m_isServer = false;
    bool m_isClient = false;
    QString m_scrollToTarget;
    QString m_checksum;
    QString m_cachedChecksum;
    QString m_backupFile;
    bool m_absentsPresent = false;
    bool m_singlesPresent = false;
    bool m_focusPlayersName = false;
    bool m_requestingRegistration = false;

};

#endif // REGISTRATIONPAGE_H
