// SPDX-FileCopyrightText: 2010-2022 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "Database.h"
#include "SharedObjects.h"
#include "StringTools.h"
#include "TournamentSettings.h"

#include "shared/Tournament.h"
#include "shared/Urls.h"

// Qt includes
#include <QDebug>
#include <QDir>
#include <QSqlError>
#include <QLockFile>

// C++ includes
#include <utility>

// The SQLite application_id for Muckurnier
// The hex representation of this number is 0x4d75636b, which maps to "Muck"
static constexpr int s_applicationId = 1299538795;

// The file format identifier for the DB_IDENTIFIER config key
static const QString s_dbIdentifier(QLatin1String("Muckturnier tournament database"));

// The revision we're using
static constexpr int s_dbVersion = 10;

// Key literals for the database's "config" table
static const QString s_keyDbIdentifier   (QLatin1String("DB_IDENTIFIER"));
static const QString s_keyDbVersion      (QLatin1String("DB_VERSION"));
static const QString s_keySettings       (QLatin1String("SETTINGS"));
static const QString s_keySettingsVersion(QLatin1String("SETTINGS_VERSION"));

// This one is only used to check if we can write to the database in a transaction that
// is rolled back afterwards. So this key will never be actually written to a database.
static const QString s_keyWriteTest(QLatin1String("WRITE_TEST"));

// Internally used connection name
static const QString s_connectionName(QLatin1String("SQLiteDB"));

// This will map to SQLite's NULL value
static const QVariant s_sqliteNull(QVariant::Int);

Database::Database(QObject *parent, SharedObjects *sharedObjects, const QString *namesSeparator)
    : QObject(parent),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_namesSeparator(namesSeparator)
{
}

// =====================
// SQL wrapper functions
// =====================

void Database::emitSqlError(const QString &text)
{
    if (! m_emitSqlErrors) {
        return;
    }

    // This prevents saveSettings() to do a write attempt when triggered by the MainWindow
    // closing the database connection after displaying an error message
    m_isWritable = false;

    emit sqlError(text);
}

bool Database::startTransaction()
{
    const bool success = m_db.transaction();
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlDatabase::transaction</p>")
                     + sqlErrorText());
        throw false;
    }
    return success;
}

bool Database::commitTransaction()
{
    const bool success = m_db.commit();
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlDatabase::commit</p>")
                     + sqlErrorText());
        throw false;
    }
    return success;
}

bool Database::rollbackTransaction(QSqlQuery &query)
{
    query.clear();
    return m_db.rollback();
}

bool Database::queryPrepare(QSqlQuery &query, const QString &text)
{
    const bool success = query.prepare(text);
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlQuery::prepare</p>")
                     + sqlErrorText(query));
        throw false;
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query)
{
    const bool success = query.exec();
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlQuery::exec</p>")
                    + sqlErrorText(query));
        throw false;
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query, const QString &text)
{
    const bool success = query.exec(text);
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlQuery::exec</p>")
                     + sqlErrorText(query));
        throw false;
    }
    return success;
}

bool Database::queryExecBatch(QSqlQuery &query)
{
    const bool success = query.execBatch();
    if (! success) {
        emitSqlError(tr("<p><b>Fehlgeschlagene Funktion:</b><br/>QSqlQuery::execBatch</p>")
                     + sqlErrorText(query));
        throw false;
    }
    return success;
}

QString Database::sqlErrorText() const
{
    const QSqlError error = m_db.lastError();
    return tr("<p><b>Die Fehlermeldung war:</b><br/>%1<br/>%2</p>").arg(
              error.driverText(), error.databaseText());
}

QString Database::sqlErrorText(const QSqlQuery &query) const
{
    const QSqlError error = query.lastError();
    QString message = tr("<p><b>Die Fehlermeldung war:</b><br/>%1<br/>%2</p>").arg(
                         error.driverText(), error.databaseText());
    const QString executedQuery = query.executedQuery();
    if (! executedQuery.isEmpty()) {
        message.append(tr("<p><b>Die ausgeführte Datenbankabfrage war:</b><br/>%3</p>").arg(
                          executedQuery));
    }
    return message;
}

// ================================
// Database connecting and checking
// ================================

bool Database::connectDatabase(const QString &dbFile)
{
    // First close a connection if we have one
    closeDatabaseConnection();

    // Connect the database
    m_db = QSqlDatabase::addDatabase(QStringLiteral("QSQLITE"), s_connectionName);
    m_db.setDatabaseName(dbFile);

    updateDbFileInfo();

    if (! m_db.open()) {
        // This should not happen
        QSqlError error = m_db.lastError();
        m_errorMessage = tr("<p>Unerwarteter Fehler: Konnte keine Datenbankverbindung herstellen."
                            "</p>"
                            "<p>Die Fehlermeldung war: %2</p>").arg(error.text());
        return false;
    }

    return true;
}

void Database::closeDatabaseConnection()
{
    if (! m_db.isOpen()) {
        return;
    }

    // Close the existing connection
    m_db.close();
    m_db = QSqlDatabase::database();
    QSqlDatabase::removeDatabase(s_connectionName);

    // Remove the lock file
    if (m_lockFile != nullptr) {
        delete m_lockFile;
        m_lockFile = nullptr;
    }

    // Clear cached variables
    m_dbFileName.clear();
    m_dbFilePath.clear();
    m_tournamentStarted = false;

    // Don't emit SQL error signals
    m_emitSqlErrors = false;
}

QString Database::lockFileName(const QString &dbFile) const
{
    const QFileInfo info(dbFile);
    QString path = info.path();
    if (path == QStringLiteral("/")) {
        path.clear();
    }
    return QStringLiteral("%1/.muckturnier.%2.lockfile").arg(path, info.fileName());
}

bool Database::lockDatabase()
{
    m_lockFile = new QLockFile(lockFileName(m_dbFileName));
    m_lockFile->setStaleLockTime(0);

    if (! m_lockFile->tryLock(0)) {
        delete m_lockFile;
        m_lockFile = nullptr;
        return false;
    }

    return true;
}

QString Database::checkForLockFile(const QString &dbFile) const
{
    // If the file doesn't exist, it can't be locked
    QFileInfo fileInfo(dbFile);
    if (! fileInfo.exists()) {
        return QString();
    }

    // If it exists, try to lock it
    const QString fileName(lockFileName(dbFile));
    QLockFile lockFile(fileName);
    lockFile.setStaleLockTime(0);
    const bool noLock = lockFile.tryLock(0);
    return noLock ? QString() : fileName;
}

void Database::updateDbFileInfo()
{
    m_dbFileName = m_db.databaseName();
    QFileInfo info(m_dbFileName);
    m_dbFilePath = info.canonicalPath();
}

Database::Error Database::openDatabase(const QString &dbFile)
{
    // Do a check if we actually do have a Muckturnier Tournament Database and if it's the correct
    // revision. If we need to do an update ot the database is not usable, we will return here.
    const Error testResult = checkDatabase(dbFile);
    if (testResult != Error::NoError) {
        return testResult;
    }

    QSqlQuery query(m_db);

    // Get the settings revision
    try {
        queryPrepare(query, QStringLiteral("SELECT value FROM config WHERE key = ?"));
        query.bindValue(0, s_keySettingsVersion);
        queryExec(query);
        if (! query.next()) {
            m_errorMessage = tr("<p>Die Revision der Einstellungen konnte nicht ausgelesen werden. "
                                "Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                                "unsachgemäß manuell verändert.</p>").arg(
                                QDir::toNativeSeparators(dbFile));
            closeDatabaseConnection();
            return Error::CouldNotOpen;
        }
        m_tournamentSettings->setDbTournamentSettingsVersion(query.value(0).toInt());
    } catch (...) {
        m_errorMessage = tr("<p>Die Revision der Einstellungen konnte nicht ausgelesen werden. "
                            "Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                            "unsachgemäß manuell verändert.</p>").arg(
                            QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // Get the settings
    QString settings;
    try {
        queryPrepare(query, QStringLiteral("SELECT value FROM config WHERE key = ?"));
        query.bindValue(0, s_keySettings);
        queryExec(query);
        query.next();
        settings = query.value(0).toString();
    } catch (...) {
        m_errorMessage = tr("<p>Die Einstellungen konnten nicht ausgelesen werden. Entweder ist "
                            "die Datenbank „%1“ beschädigt, oder sie wurde unsachgemäß manuell "
                            "verändert.</p>").arg(QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    if (! m_tournamentSettings->load(settings)) {
        m_errorMessage = tr("<p>Beim Parsen der Turniereinstellungen ist ein Fehler aufgetreten. "
                            "Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                            "unsachgemäß manuell verändert.</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // Initialize the players count, scores per round and the rounds state

    try {
        queryExec(query, QStringLiteral("SELECT COUNT(id) FROM players"));
        query.next();
        setPairsOrPlayersCount(query.value(0).toInt());
    } catch (...) {
        m_errorMessage = tr("<p>Beim Zählen der Anmeldungen ist ein Fehler aufgetreten. "
                            "Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                            "unsachgemäß manuell verändert.</p>").arg(
                            QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    if (! updateRoundsState()) {
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    if (! checkIfTournamentStarted()) {
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // From here on, not a single SQL query should fail, so we activate the MainWindow's SQL
    // error handling (which displays a critical error and closes the database if something goes
    // wrong)
    m_emitSqlErrors = true;
    return Error::NoError;
}

Database::Error Database::checkDatabase(const QString &dbFile)
{
    // Check the given file name to be an actual file and readable.
    // No interaction with the database itself is done yet here.

    // Check if it's a file and if we can read it
    const QFileInfo fileInfo = QFileInfo(dbFile);
    if (! fileInfo.exists()) {
        m_errorMessage = tr("<p>Die Datei „%1“ konnte nicht gefunden werden.</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        m_errorMessage = tr("<p>„%1“ ist weder eine Datei, noch ein symbolischer Link.</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! fileInfo.isReadable()) {
        m_errorMessage = tr("<p>Die Datei „%1“ ist nicht lesbar. Bitte die Zugriffsrechte "
                            "überprüfen!</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }

    // Try to establish a connection to the file
    if (! connectDatabase(dbFile)) {
        m_errorMessage = tr("<p>Es konnte keine Datenbankverbindung zu der Datei „%1“ aufgebaut "
                            "werden.</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }

    // We could connect to the file

    // Let's see if we can acquire a lock (before starting a normal transaction)
    const bool dbIsLocked = checkIfDbIsLocked();

    // First we try if we can start a transaction (needed for the write test later) ...
    if (! m_db.transaction()) {
        m_errorMessage = tr("<p>Für die Datenbank „%1“ konnte keine Transaktion gestartet werden. "
                            "Das sollte nicht passieren. Die Datenbank kann nicht verwendet "
                            "werden!</p>").arg(QDir::toNativeSeparators(dbFile));
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // ... then we create a query
    QSqlQuery query(m_db);

    // Do an internal check to see if it's an SQLite database and if it's in a sane condition
    try {
        queryExec(query, QStringLiteral("PRAGMA quick_check"));
        if (! query.next() || query.value(0).toString() != QLatin1String("ok")) {
            throw false;
        }
    } catch (...) {
        m_errorMessage = tr("<p>Der Datenbankcheck für die Datei „%1“ ist fehlgeschlagen.</p>").arg(
                            QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // Check if we actually have a Muckturnier tournament database

    // Let's see if we have a valid database revision config entry

    int dbVersion = 0;
    try {
        queryPrepare(query, QStringLiteral("SELECT value FROM config WHERE key = ?"));
        query.bindValue(0, s_keyDbVersion);
        queryExec(query);
        if (! query.next()) {
            m_errorMessage = tr("<p>Die verwendete Datenbankrevision konnte nicht ausgelesen "
                                "werden.</p>"
                                "<p>Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                                "unsachgemäß manuell verändert.</p>").arg(
                                QDir::toNativeSeparators(dbFile));
            closeDatabaseConnection();
            return Error::CouldNotOpen;
        }
        dbVersion = query.value(0).toInt();
    } catch (...) {
        m_errorMessage = tr("<p>Die verwendete Datenbankrevision konnte nicht ausgelesen werden."
                            "</p>"
                            "<p>Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                            "unsachgemäß manuell verändert.</p>").arg(
                            QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    if (dbVersion < 3) {
        m_errorMessage = tr("<p>Die von der Datenbank „%1“ benutzte Revision %2 ist unbekannt und "
                            "kann nicht verarbeitet werden.</p>").arg(
                            QDir::toNativeSeparators(dbFile), QString::number(dbVersion));
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    } else if (dbVersion > s_dbVersion) {
        m_errorMessage = tr("<p>Die von der Datenbank „%1“ benutzte Revision %2 ist unbekannt und "
                            "kann nicht verarbeitet werden.</p>"
                            "<p>Vermutlich wurde die Datei mit einer neueren Version von "
                            "Muckturnier erstellt. Bitte <a href=\"%3\">aktualisieren</a>!").arg(
                            QDir::toNativeSeparators(dbFile),
                            QString::number(dbVersion),
                            Urls::download);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // If the revision is between 3 and s_dbVersion, we should have a valid Muckturnier tournament
    // database. Let's see if this is really the case

    // Starting with dbv8, we have a (most probably) unique identifier, which we check first.
    if (dbVersion >= 8) {
        bool identifierOkay = false;
        try {
            queryPrepare(query, QStringLiteral("SELECT value FROM config WHERE key = ?"));
            query.bindValue(0, s_keyDbIdentifier);
            queryExec(query);
            if (! query.next()) {
                m_errorMessage = tr("<p>Die Datenbank-ID ist nicht auffindbar. Entweder ist die "
                                    "Datenbank „%1“ beschädigt, oder sie wurde unsachgemäß manuell "
                                    "verändert.</p>").arg(QDir::toNativeSeparators(dbFile));
                closeDatabaseConnection();
                return Error::CouldNotOpen;
            }
            identifierOkay = query.value(0).toString() == s_dbIdentifier;
        } catch (...) {
            m_errorMessage = tr("<p>Die Datenbank-ID konnte nicht ausgelesen werden. Entweder ist "
                                "die Datenbank „%1“ beschädigt, oder sie wurde unsachgemäß manuell "
                                "verändert.</p>").arg(QDir::toNativeSeparators(dbFile))
                             + sqlErrorText(query);
            closeDatabaseConnection();
            return Error::CouldNotOpen;
        }

        if (! identifierOkay) {
            m_errorMessage = tr("<p>Die Datenbank „%1“ hat keine gültige Datenbank-ID. Vermutlich "
                                "ist es keine Muckturnier-Turnierdatenbank.</p>").arg(
                                QDir::toNativeSeparators(dbFile));
            closeDatabaseConnection();
            return Error::CouldNotOpen;
        }

    // For older revisions, we have to rely on the old structure check
    // where we simply look for all tables to be present.
    } else {
        const QStringList tables = m_db.tables();
        bool allTablesPresent = false;

        if (dbVersion >= 7) {
            allTablesPresent =     tables.contains(QLatin1String("config"))
                                && tables.contains(QLatin1String("players"))
                                && tables.contains(QLatin1String("markers"))
                                && tables.contains(QLatin1String("assignment"))
                                && tables.contains(QLatin1String("score"));
        } else {
            // The "markers" table wasn't present until dbv7
            allTablesPresent =     tables.contains(QLatin1String("config"))
                                && tables.contains(QLatin1String("players"))
                                && tables.contains(QLatin1String("assignment"))
                                && tables.contains(QLatin1String("score"));
        }

        if (! allTablesPresent) {
            m_errorMessage = tr("<p>Die Datenbank „%1“ enthält nicht alle nötigen Tabellen.</p>"
                                "<p>Entweder ist es keine Muckturnier-Turnierdatenbank, die "
                                "Datenbank ist beschädigt oder sie wurde unsachgemäß manuell "
                                "verändert.</p>").arg(QDir::toNativeSeparators(dbFile));
            closeDatabaseConnection();
            return Error::CouldNotOpen;
        }
    }

    // Check if all configuration keys other than DB_VERSION and DB_IDENTIFIER are present
    // (those have already been checked to be present above)

    QString remainingConfigKeys;
    int expectedRemainingKeysCount = 0;

    if (dbVersion >= 3 && dbVersion <= 4) {
        expectedRemainingKeysCount = 4;
        remainingConfigKeys = QStringLiteral("'MUCKTURNIER_VERSION',"
                                             "'TOURNAMENT_MODE',"
                                             "'BOOGER_SCORE',"
                                             "'GAMES_PER_ROUND'");
    } else if (dbVersion == 5) {
        expectedRemainingKeysCount = 4;
        remainingConfigKeys = QStringLiteral("'MUCKTURNIER_VERSION',"
                                             "'TOURNAMENT_MODE',"
                                             "'BOOGER_SCORE',"
                                             "'BOOGERS_PER_ROUND'");
    } else if (dbVersion >= 6 && dbVersion <= 7) {
        expectedRemainingKeysCount = 7;
        remainingConfigKeys = QStringLiteral("'MUCKTURNIER_VERSION',"
                                             "'TOURNAMENT_MODE',"
                                             "'BOOGER_SCORE',"
                                             "'BOOGERS_PER_ROUND',"
                                             "'NO_TABLE_NUMBERS',"
                                             "'AUTO_SELECT_PAIRS',"
                                             "'INCLUDE_OPPONENT_GOALS'");
    } else if (dbVersion == 8) {
        expectedRemainingKeysCount = 1;
        remainingConfigKeys = QStringLiteral("'SETTINGS'");
    } else { // dbVersion >= 9
        expectedRemainingKeysCount = 2;
        remainingConfigKeys = QStringLiteral("'SETTINGS',"
                                             "'SETTINGS_VERSION'");
    }

    int foundKeysCount = 0;
    try {
        queryExec(query, QStringLiteral("SELECT COUNT(key) FROM config WHERE key IN(%1)").arg(
                                        remainingConfigKeys));
        query.next();
        foundKeysCount = query.value(0).toInt();
    } catch (...) {
        m_errorMessage = tr("<p>Die Konfigurationsschlüssel konnten nicht ausgelesen werden.</p>"
                            "<p>Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                            "unsachgemäß manuell verändert.</p>").arg(
                            QDir::toNativeSeparators(dbFile))
                         + sqlErrorText(query);
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    if (foundKeysCount != expectedRemainingKeysCount) {
        m_errorMessage = tr("<p>Die Datenbank „%1“ enthält nicht alle nötigen "
                            "Konfigurationsschlüssel.</p>"
                            "<p>Entweder ist es keine Muckturnier-Turnierdatenbank, die "
                            "Datenbank ist beschädigt oder sie wurde unsachgemäß manuell "
                            "verändert.</p>").arg(QDir::toNativeSeparators(dbFile));
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // Let's see if we can write to the database.
    //
    // We do have to check if a database is writable by actually trying to insert some data (is
    // this a bug or a feature?!) because:
    //
    // - If the file itself has write access, QFileInfo::isWritable() will return true, but if the
    //   parent directory is not writable, we won't be able to write to the database.
    //
    // - Checking for the parent directory to be writable in addition does also not help, as we will
    //   get false for "My Documents" on Windows, although we can write to it.
    //
    // - Opening the file via QFile::open() does actually work if the file is writable and the
    //   parent directory is not (with the QIODevice::WriteOnly flag, the file will even be
    //   trucated!), but we still can't insert data in this case.
    //
    // So here we go:

    try {
        // Even if the database is locked by another instance, it's save to do this here:
        // The transaction won't be actually written to the database. And even if it was,
        // it wouldn't affect the other instance, because s_keyWriteTest is never used.
        // So this is save in each case:
        queryPrepare(query, QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
        query.bindValue(0, s_keyWriteTest);
        query.bindValue(1, QStringLiteral("test"));
        queryExec(query);
        m_isWritable = true;
    } catch (...) {
        m_isWritable = false;
    }

    // Here, we don't interact with the database anymore, so we can rollback the transaction
    // so that no changes will be actually written to the database
    rollbackTransaction(query);

    // If the database is writable, we have to check if it's locked,
    // either from SQLite itself, or by another instance.

    if (m_isWritable) {
        if (dbIsLocked) {
            closeDatabaseConnection();
            m_errorMessage = tr("<p><b>Unerwarteter Fehler:</b> Die Datenbank ist gesperrt</p>"
                                "<p>Liegt die Datenbank auf einer Windows-Netzwerk-Freigabe (bzw. "
                                "einem Samba-Share/CIFS-Mount)? Das Öffnen einer solchen Datenbank "
                                "ist nur mit dateisystemseitigem Schreibschutz möglich.</p>"
                                "<p>Ansonsten sollte das nicht passieren, sofern kein anderer "
                                "Prozess die Datenbank gesperrt hat (und das sollte eigentlich "
                                "auch nicht der Fall sein)!");
            return Error::CouldNotOpen;
        }

        // If the database is writable, we have to make sure we're the only instance using it
        if (! lockDatabase()) {
            closeDatabaseConnection();
            m_errorMessage = tr("<p>Die Datenbank ist bereits von einer anderen Instanz von "
                                "Muckturnier geöffnet.</p>"
                                "<p>Sollte dies nicht der Fall sein (evtl. nach einen Crash, "
                                "Stromausfall etc.), dann bitte die Lock-Datei „%1“ manuell "
                                "löschen und die Datenbank erneut öffnen.</p>").arg(
                                QDir::toNativeSeparators(lockFileName(dbFile)));
            return Error::CouldNotOpen;
        }
    }

    // Check the database revision
    if (dbVersion == 3) {
        return Error::DbVersion3;
    } else if (dbVersion == 4) {
        return Error::DbVersion4;
    } else if (dbVersion == 5) {
        return Error::DbVersion5;
    } else if (dbVersion == 6) {
        return Error::DbVersion6;
    } else if (dbVersion == 7) {
        return Error::DbVersion7;
    } else if (dbVersion == 8) {
        return Error::DbVersion8;
    } else if (dbVersion == 9) {
        return Error::DbVersion9;
    } else if (dbVersion != s_dbVersion) {
        // This should not happen, as long as we update this code block with each version bump :-P
        m_errorMessage = tr("<p>Die Datenbank „%1“ hat die falsche Revision (ist %2, die benutzte "
                            "Version von Muckturnier nutzt %3). Die Datenbank kann nicht geöffnet "
                            "werden").arg(QDir::toNativeSeparators(dbFile),
                                          QString::number(dbVersion),
                                          QString::number(s_dbVersion));
        closeDatabaseConnection();
        return Error::CouldNotOpen;
    }

    // Everything is fine :-)
    return Error::NoError;
}

// Functions using SQL, used while opening/creating a database
// ===========================================================

bool Database::checkIfDbIsLocked()
{
    QSqlQuery query(m_db);
    if (query.exec(QStringLiteral("BEGIN EXCLUSIVE"))) {
        query.exec(QStringLiteral("COMMIT"));
        return false;
    } else {
        return true;
    }
}

bool Database::create(const QString &dbFile)
{
    if (! connectDatabase(dbFile)) {
        return false;
    }

    if (checkIfDbIsLocked()) {
        closeDatabaseConnection();
        m_errorMessage = tr("<p><b>Unerwarteter Fehler:</b> Die Datenbank ist gesperrt</p>"
                            "<p>Das tritt auf, wenn eine Datenbank auf einer Windows-Netzwerk-"
                            "Freigabe (bzw. einem Samba-Share/CIFS-Mount) angelegt werden soll; "
                            "das ist nicht möglich.</p>"
                            "<p>Ansonsten sollte das nicht passieren, sofern kein anderer Prozess "
                            "die Datenbank gesperrt hat!");
        return false;
    }

    if (! lockDatabase()) {
        closeDatabaseConnection();
        m_errorMessage = tr("<p><b>Unerwarteter Fehler:</b> Die Lock-Datei „%1“ konnte nicht "
                            "erstellt werden. Bitte die Zugriffsrechte des Verzeichnisses „%2“ "
                            "überprüfen!</p>").arg(
                            QDir::toNativeSeparators(m_dbFilePath),
                            QDir::toNativeSeparators(lockFileName(m_dbFileName)));
        return false;
    }

    if (! startTransaction()) {
        return false;
    }

    QSqlQuery query(m_db);

    try {
        // Save the custom header fields
        // Using bind values in a PRAGMA call doesn't work, so we have to use a literal query
        queryExec(query, QStringLiteral("PRAGMA application_id = %1").arg(s_applicationId));
        queryExec(query, QStringLiteral("PRAGMA user_version = %1").arg(s_dbVersion));

        // Setup the tables
        queryExec(query, QStringLiteral("CREATE TABLE config(\n"
                                        "    key TEXT,\n"
                                        "    value TEXT\n"
                                        ")"));
        queryExec(query, QStringLiteral("CREATE TABLE players(\n"
                                        "    id INTEGER PRIMARY KEY,\n"
                                        "    name TEXT,\n"
                                        "    marker INTEGER,\n"
                                        "    disqualified INTEGER\n"
                                        ")"));
        queryExec(query, QStringLiteral("CREATE TABLE markers(\n"
                                        "    id INTEGER PRIMARY KEY,\n"
                                        "    name TEXT,\n"
                                        "    color INTEGER,\n"
                                        "    sorting INTEGER,\n"
                                        "    sequence INTEGER\n"
                                        ")"));
        queryExec(query, QStringLiteral("CREATE TABLE assignment(\n"
                                        "    round INTEGER,\n"
                                        "    pair_or_player INTEGER,\n"
                                        "    pair_number INTEGER,\n"
                                        "    player_number INTEGER,\n"
                                        "    table_number INTEGER\n"
                                        ")"));
        queryExec(query, QStringLiteral("CREATE TABLE score(\n"
                                        "    round INTEGER,\n"
                                        "    table_number INTEGER,\n"
                                        "    booger INTEGER,\n"
                                        "    pair_or_player INTEGER,\n"
                                        "    score INTEGER\n"
                                        ")"));

        // Add the draw cleanup trigger
        queryExec(query, QStringLiteral(
            "CREATE TRIGGER cleanup_draw AFTER DELETE ON players\n"
            "FOR EACH ROW\n"
            "BEGIN\n"
            "    DELETE FROM assignment WHERE pair_or_player = OLD.id;\n"
            "END"));

        // Insert the default config keys and values
        queryPrepare(query, QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
        query.addBindValue(QVariantList { s_keyDbIdentifier,
                                          s_keyDbVersion,
                                          s_keySettings,
                                          s_keySettingsVersion });
        query.addBindValue(QVariantList { s_dbIdentifier,
                                          s_dbVersion,
                                          s_sqliteNull,
                                          m_tournamentSettings->version() });
        queryExecBatch(query);

        commitTransaction();

    } catch (...) {
        rollbackTransaction(query);
        m_errorMessage = tr("<p>Beim Erstellen der Datenbankstruktur ist ein Fehler aufgetreten. "
                            "Das sollte nicht passieren! Evtl. wurde die Datei während des "
                            "Erstellens von einem anderen Prozess verändert?</p>")
                         + sqlErrorText(query);
        m_isWritable = false;
        return false;
    }

    updateDbFileInfo();
    m_pairsOrPlayersCount = 0;
    m_scoresPerRound = 0;
    m_roundsState = QMap<int, bool> { { 1, false } };
    m_isWritable = true;
    m_tournamentStarted = false;
    m_emitSqlErrors = true;

    return true;
}

bool Database::checkIfTournamentStarted()
{
    QSqlQuery query(m_db);
    try {
        queryExec(query, QStringLiteral("SELECT COUNT(round) FROM score"));
        query.next();
        m_tournamentStarted = query.value(0).toInt() != 0;
        return true;
    } catch (...) {
        if (! m_emitSqlErrors) {
            m_errorMessage = tr("<p>Beim Zählen der Ergebnisse ist ein Fehler aufgetreten. "
                                "Entweder ist die Datenbank „%1“ beschädigt, oder sie wurde "
                                "unsachgemäß manuell verändert.</p>").arg(
                                QDir::toNativeSeparators(m_dbFileName))
                             + sqlErrorText(query);
        }
        return false;
    }
}

bool Database::updateRoundsState()
{
    m_roundsState.clear();
    QSqlQuery query(m_db);

    try {
        queryExec(query, QStringLiteral("SELECT round, COUNT(round) FROM score GROUP BY round"));
        while (query.next()) {
            m_roundsState[query.value(0).toInt()] = query.value(1).toInt() == m_scoresPerRound;
        }
    } catch (...) {
        if (! m_emitSqlErrors) {
            m_errorMessage = tr("<p>Beim Aktualisieren des Rundenstatus ist ein Fehler "
                                "aufgetreten. Entweder ist die Datenbank „%1“ beschädigt, oder "
                                "sie wurde unsachgemäß manuell verändert.</p>").arg(
                                QDir::toNativeSeparators(m_dbFileName))
                             + sqlErrorText(query);
        }
        return false;
    }

    // Be sure to have a state for each round. All data from a round could have been deleted.
    // If we have no data, round 1 will be added as open.
    const int lastRound = m_roundsState.isEmpty() ? 1 : m_roundsState.lastKey();
    for (int i = 1; i <= lastRound; i++) {
        if (! m_roundsState.contains(i)) {
            m_roundsState[i] = false;
        }
    }

    m_allRounds = m_roundsState.count();

    return true;
}

// Database revision update functions
// ==================================

bool Database::updateDatabase(const QString &dbFileName, Error dbVersion)
{
    if (! connectDatabase(dbFileName) || ! startTransaction()) {
        return false;
    }

    QSqlQuery query(m_db);

    try {
        switch (dbVersion) {
        case NoError:
        case CouldNotOpen:
            Q_UNREACHABLE();
            break;
        case Error::DbVersion3:
            updateDatabaseFromV3(query);
            [[fallthrough]];
        case Error::DbVersion4:
            updateDatabaseFromV4(query);
            [[fallthrough]];
        case Error::DbVersion5:
            updateDatabaseFromV5(query);
            [[fallthrough]];
        case Error::DbVersion6:
            updateDatabaseFromV6(query);
            [[fallthrough]];
        case Error::DbVersion7:
            updateDatabaseFromV7(query);
            [[fallthrough]];
        case Error::DbVersion8:
            updateDatabaseFromV8(query);
            [[fallthrough]];
        case Error::DbVersion9:
            updateDatabaseFromV9(query);
        }

        commitTransaction();
        closeDatabaseConnection();

    } catch (...) {
        rollbackTransaction(query);
        closeDatabaseConnection();
        m_errorMessage = sqlErrorText(query);
        m_isWritable = false;
        return false;
    }

    return true;
}

void Database::updateDatabaseFromV3(QSqlQuery &query)
{
    // Get the tournament mode
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    const QString tournamentMode = query.value(0).toString();

    // Add the missing column to "assignment"
    queryExec(query, QStringLiteral("ALTER TABLE assignment ADD COLUMN player_number INTEGER"));

    // Fill in the data. For a single player mode database, the fetch result hopefully is still in
    // the correct order, so that the right players become player 1 and 2 inside a team.
    if (tournamentMode == QStringLiteral("FIXED_PAIRS")) {
        queryExec(query, QStringLiteral("UPDATE assignment SET player_number = 1"));
    } else {
        QVariantList round;
        QVariantList pair_or_player;
        QVariantList pair_number;
        QVariantList table_number;
        QVariantList player_number;
        queryExec(query, QStringLiteral("SELECT round, pair_or_player, pair_number, table_number "
                                        "FROM assignment "
                                        "ORDER BY round, table_number, pair_number"));
        int playerNumber = 1;
        while (query.next()) {
            round.append(query.value(0).toInt());
            pair_or_player.append(query.value(1).toInt());
            pair_number.append(query.value(2).toInt());
            table_number.append(query.value(3).toInt());
            playerNumber = (playerNumber + 1) % 2;
            player_number.append(playerNumber + 1);
        }

        queryPrepare(query, QStringLiteral("UPDATE assignment SET player_number = ? "
                                           "WHERE round = ? "
                                           "AND pair_or_player = ? "
                                           "AND pair_number = ? "
                                           "AND table_number = ?"));
        query.addBindValue(player_number);
        query.addBindValue(round);
        query.addBindValue(pair_or_player);
        query.addBindValue(pair_number);
        query.addBindValue(table_number);
        queryExecBatch(query);
    }

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 4 WHERE key = 'DB_VERSION'"));
}

void Database::updateDatabaseFromV4(QSqlQuery &query)
{
    // Update "GAMES_PER_ROUND" to "BOOGERS_PER_ROUND"
    queryExec(query, QStringLiteral("UPDATE config SET key = 'BOOGERS_PER_ROUND' "
                                                "WHERE key = 'GAMES_PER_ROUND'"));

    // Update the score table's "game" column to "booger"
    queryExec(query, QStringLiteral("ALTER TABLE score RENAME TO score_tmp"));
    queryExec(query, QStringLiteral("CREATE TABLE score(\n"
                                    "    round INTEGER,\n"
                                    "    table_number INTEGER,\n"
                                    "    booger INTEGER,\n"
                                    "    pair_or_player INTEGER,\n"
                                    "    score INTEGER\n"
                                    ")"));
    queryExec(query, QStringLiteral("INSERT INTO score(round, table_number, booger, "
                                                      "pair_or_player, score) "
                                    "SELECT round, table_number, game, pair_or_player, score "
                                    "FROM score_tmp"));
    queryExec(query, QStringLiteral("DROP TABLE score_tmp"));

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 5 WHERE key = 'DB_VERSION'"));
}

void Database::updateDatabaseFromV5(QSqlQuery &query)
{
    // The value for the new AUTO_SELECT_PAIRS setting depends on the tournament mode. This option
    // has always been on by default for fixed pairs, but must not be set for single player
    // tournaments, as it's not applicable in this mode (and also not changeable, as the respective
    // checkbox will be hidden)
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    const QString autoSelectPairs = query.value(0).toString() == QLatin1String("FIXED_PAIRS")
                                                                     ? QStringLiteral("1")
                                                                     : QStringLiteral("0");

    // Insert the new config keys. We don't use the defined constants here in case they change with
    // some later version.
    queryPrepare(query, QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.addBindValue(QVariantList { QStringLiteral("NO_TABLE_NUMBERS"),
                                      QStringLiteral("AUTO_SELECT_PAIRS"),
                                      QStringLiteral("INCLUDE_OPPONENT_GOALS") });
    query.addBindValue(QVariantList { QStringLiteral("0"),
                                      autoSelectPairs,
                                      QStringLiteral("0") });
    queryExecBatch(query);

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 6 WHERE key = 'DB_VERSION'"));
}

void Database::updateDatabaseFromV6(QSqlQuery &query)
{
    // Add the new "markers" table
    queryExec(query, QStringLiteral("CREATE TABLE markers(\n"
                                    "    id INTEGER PRIMARY KEY,\n"
                                    "    name TEXT,\n"
                                    "    color INTEGER,\n"
                                    "    sorting INTEGER,\n"
                                    "    sequence INTEGER\n"
                                    ")"));

    // Add the new "marker" column to the "players" table
    queryExec(query, QStringLiteral("ALTER TABLE players ADD COLUMN marker INTEGER"));

    // Add the default markers using a direct query in case the format changes later:
    queryPrepare(query, QStringLiteral("INSERT INTO markers(id, name, color, sorting, sequence) "
                                       "VALUES(?, ?, ?, ?, ?)"));
    query.addBindValue(QVariantList { 0, 1, 2, 3 });
    query.addBindValue(QVariantList { QString(),
                                      tr("Rot"),
                                      tr("Blau"),
                                      tr("Grün") });
    // These are the output of QColor(0,   0,   0,   0).rgb()
    //                         QColor(255, 0,   0,   0).rgb()
    //                         QColor(0,   0,   255, 0).rgb()
    //                         QColor(0,   128, 0,   0).rgb()
    query.addBindValue(QVariantList { (uint) 4278190080,
                                      (uint) 4294901760,
                                      (uint) 4278190335,
                                      (uint) 4278222848 });
    query.addBindValue(QVariantList { 0, 0, 0, 1 });
    query.addBindValue(QVariantList { 2, 0, 1, 3 });
    queryExecBatch(query);

    // Map the old markers to the new ones

    QVariantList oldNames;
    QVariantList newNames;
    QVariantList markerIds;

    // Search for marked names
    const QString markerCharacter = QStringLiteral("\u200B");
    queryPrepare(query, QStringLiteral("SELECT name FROM players"));
    queryExec(query);
    while (query.next()) {
        const QString name = query.value(0).toString();
        int marker = -1;
        if (name.startsWith(markerCharacter) && ! name.endsWith(markerCharacter)) {
            marker = 1;
        } else if (! name.startsWith(markerCharacter) && name.endsWith(markerCharacter)) {
            marker = 3;
        }
        if (marker == -1) {
            continue;
        }
        QString strippedName = name;
        strippedName.remove(markerCharacter);
        oldNames.append(name);
        newNames.append(strippedName);
        markerIds.append(marker);
    }

    // Update the names and the marker ids
    queryPrepare(query, QStringLiteral("UPDATE players SET name = ?, marker = ? WHERE name = ?"));
    query.addBindValue(newNames);
    query.addBindValue(markerIds);
    query.addBindValue(oldNames);
    queryExecBatch(query);

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 7 WHERE key = 'DB_VERSION'"));
}

void Database::updateDatabaseFromV7(QSqlQuery &query)
{
    // Add the preassignment cleanup trigger
    queryExec(query, QStringLiteral(
        "CREATE TRIGGER cleanup_preassignment AFTER DELETE ON players\n"
        "FOR EACH ROW\n"
        "BEGIN\n"
        "    DELETE FROM assignment WHERE pair_or_player = OLD.id;\n"
        "END"));

    // Move the settings to the new JSON format
    QString jsonSettings(QStringLiteral("{"));

    // We have to fix a bug introduced with version 0.7.7 when updating from dbv6 or dbv7 (when
    // updating a dbv5 database, it will be fixed directly in updateDatabaseFromV5 now).
    // If a dbv5 database was updated with v0.7.7 up to v3.2, AUTO_SELECT_PAIRS was erroneously set
    // to TRUE for single player databases and not deselected later anymore, which lead to wrong
    // warnings or complete unusability of the score page when selecting players 1.

    // First, we need the tournament mode. It's selected again below, but we need it when processing
    // the AUTO_SELECT_PAIRS key. So here we go:
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    const bool singlePlayerMode = query.value(0).toString() == QLatin1String("SINGLE_PLAYERS");

    // Convert the existing keys and values
    queryExec(query, QStringLiteral("SELECT key, value FROM config"));
    while (query.next()) {
        const QString key(query.value(0).toString());
        const QString value(query.value(1).toString());

        if (key == QStringLiteral("MUCKTURNIER_VERSION")) {
            jsonSettings.append(QLatin1String("\"muckturnier version\":\""));
            jsonSettings.append(value);
            jsonSettings.append(QLatin1String("\","));
        } else if (key == QStringLiteral("TOURNAMENT_MODE")) {
            jsonSettings.append(QLatin1String("\"tournament mode\":\""));
            jsonSettings.append(value == QStringLiteral("FIXED_PAIRS")
                                    ? QLatin1String("fixed pairs\",")
                                    : QLatin1String("single players\","));
        } else if (key == QStringLiteral("BOOGER_SCORE")) {
            jsonSettings.append(QLatin1String("\"booger score\":"));
            jsonSettings.append(value);
            jsonSettings.append(QLatin1String(","));
        } else if (key == QStringLiteral("BOOGERS_PER_ROUND")) {
            jsonSettings.append(QLatin1String("\"boogers per round\":"));
            jsonSettings.append(value);
            jsonSettings.append(QLatin1String(","));
        } else if (key == QStringLiteral("NO_TABLE_NUMBERS")) {
            jsonSettings.append(QLatin1String("\"no table numbers\":"));
            jsonSettings.append(value == QStringLiteral("0") ? QLatin1String("false")
                                                             : QLatin1String("true"));
            jsonSettings.append(QLatin1String(","));
        } else if (key == QStringLiteral("AUTO_SELECT_PAIRS")) {
            jsonSettings.append(QLatin1String("\"autoselect pairs\":"));
            // And here's the fix :-)
            if (singlePlayerMode) {
                jsonSettings.append(QLatin1String("false"));
            } else {
                jsonSettings.append(value == QStringLiteral("0") ? QLatin1String("false")
                                                                 : QLatin1String("true"));
            }
            jsonSettings.append(QLatin1String(","));
        } else if (key == QStringLiteral("INCLUDE_OPPONENT_GOALS")) {
            jsonSettings.append(QLatin1String("\"include opponent goals\":"));
            jsonSettings.append(value == QStringLiteral("0") ? QLatin1String("false")
                                                             : QLatin1String("true"));
            jsonSettings.append(QLatin1String(","));
        }
    }

    // Add the default new ones
    jsonSettings.append(QLatin1String("\"absent marker\":-1,"));
    jsonSettings.append(QLatin1String("\"default marker\":0,"));
    jsonSettings.append(QLatin1String("\"singles management version\":1,"));
    jsonSettings.append(QLatin1String("\"singles management\":{"
                                          "\"assigned marker\":0,"
                                          "\"condition\":0,"
                                          "\"condition string\":\"/\","
                                          "\"enabled\":false,"
                                          "\"mark singles\":false,"
                                          "\"single marker\":0},"));
    jsonSettings.append(QLatin1String("\"players page columns\":{"
                                          "\"assignment\":false,"
                                          "\"mark\":true,"
                                          "\"marker\":false}"));

    // Finish the string representation
    jsonSettings.append(QLatin1String("}"));

    // Delete the old config keys and values
    queryExec(query, QStringLiteral("DELETE FROM config"));

    // Insert the new config keys and values
    queryPrepare(query, QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.addBindValue(QVariantList {
        QStringLiteral("DB_VERSION"),
        QStringLiteral("DB_IDENTIFIER"),
        QStringLiteral("SETTINGS")
    });
    query.addBindValue(QVariantList {
        8, // <-- This updates DB_VERSION
        QStringLiteral("Muckturnier tournament database"),
        jsonSettings
    });
    queryExecBatch(query);

    // Add the new custom header fields
    queryExec(query, QStringLiteral("PRAGMA application_id = 1299538795"));
    queryExec(query, QStringLiteral("PRAGMA user_version = 8"));
}

void Database::updateDatabaseFromV8(QSqlQuery &query)
{
    // Add the SETTINGS_VERSION key to be able to track changes in the settings
    queryExec(query, QStringLiteral("INSERT INTO config(key, value) "
                                    "VALUES('SETTINGS_VERSION', 2)"));

    // Move the "players page columns" hash to the "registration columns" array
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'SETTINGS'"));
    query.next();
    const auto settings = m_tournamentSettings->updateSettingsFromDbV8(query.value(0).toString());
    query.prepare(QStringLiteral("UPDATE config SET value = ? WHERE key = 'SETTINGS'"));
    query.bindValue(0, settings);
    queryExec(query);

    // Rename the cleanup_preassignment trigger
    queryExec(query, QStringLiteral("DROP TRIGGER cleanup_preassignment"));
    queryExec(query, QStringLiteral("CREATE TRIGGER cleanup_draw AFTER DELETE ON players\n"
                                    "FOR EACH ROW\n"
                                    "BEGIN\n"
                                    "    DELETE FROM assignment WHERE pair_or_player = OLD.id;\n"
                                    "END"));

    // Update the DB_VERSION
    queryExec(query, QStringLiteral("UPDATE config SET value = 9 WHERE key = 'DB_VERSION'"));
    queryExec(query, QStringLiteral("PRAGMA user_version = 9"));
}

void Database::updateDatabaseFromV9(QSqlQuery &query)
{
    // Add the "disqualified" column to the players table
    queryExec(query, QStringLiteral("ALTER TABLE players ADD COLUMN disqualified INTEGER"));

    // Update the DB_VERSION
    queryExec(query, QStringLiteral("UPDATE config SET value = 10 WHERE key = 'DB_VERSION'"));
    queryExec(query, QStringLiteral("PRAGMA user_version = 10"));
}

// =============================================================
// Functions using SQL, used while working on an opened database
// =============================================================

// Functions reading data
// ======================

TSuccess<QStringList> Database::players()
{
    QStringList result;
    bool success;

    readHelper([this, &result](QSqlQuery &query) {
        queryExec(query, QStringLiteral("SELECT name FROM players"));
        while (query.next()) {
            result.append(query.value(0).toString());
        }
    }, success);

    return { result, success };
}

TSuccess<QHash<QString, Database::PlayersData>> Database::playersData()
{
    QHash<QString, PlayersData> result;
    bool success;

    readHelper([this, &result](QSqlQuery &query) {
        queryExec(query, QStringLiteral("SELECT "
                                            "players.name, "
                                            "players.marker, "
                                            "assignment.table_number, "
                                            "assignment.pair_number, "
                                            "players.disqualified "
                                        "FROM players "
                                        "LEFT JOIN assignment ON "
                                            "assignment.pair_or_player = players.id "
                                            "AND assignment.round = 0"));
        while (query.next()) {
            PlayersData playersData;
            playersData.marker = query.value(1).toInt();
            playersData.assignedTable = query.value(2).toInt();
            playersData.assignedPairNumber = query.value(3).toInt();
            playersData.disqualified = query.value(4).toInt();
            result.insert(query.value(0).toString(), playersData);
        }
    }, success);

    return { result, success };
}

TSuccess<QVector<int>> Database::missingTables(int round)
{
    QVector<int> result;
    bool success;

    readHelper([this, round, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("SELECT DISTINCT table_number FROM score "
                                           "WHERE round = ?"));
        query.bindValue(0, round);
        queryExec(query);

        QVector<int> tablesWithScores;
        while (query.next()) {
            tablesWithScores.append(query.value(0).toInt());
        }
        result.reserve(m_tablesCount - tablesWithScores.size());

        for (int i = 1; i <= m_tablesCount; i++) {
            if (! tablesWithScores.contains(i)) {
                result.append(i);
            }
        }
    }, success);

    return { result, success };
}

TSuccess<QHash<QString, int>> Database::missingPlayers(int round)
{
    QHash<QString, int> result;
    bool success;

    readHelper([this, round, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral(
            "SELECT players.name, players.id FROM players "
            "WHERE NOT EXISTS ("
                "SELECT score.pair_or_player FROM score "
                "WHERE players.id = score.pair_or_player "
                "AND round = ?"
            ")"));
        query.bindValue(0, round);
        queryExec(query);
        while (query.next()) {
            result.insert(query.value(0).toString(), query.value(1).toInt());
        }
    }, success);

    return { result, success };
}

TSuccess<Database::Assignment> Database::assignment(int round)
{
    Assignment result;
    bool success;

    readHelper([this, round, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral(
            "SELECT table_number, pair_number, player_number, pair_or_player "
            "FROM assignment WHERE round = ? "
            "ORDER BY table_number, pair_number, player_number, pair_or_player"));
        query.bindValue(0, round - 1);
        queryExec(query);

        QHash<int, QVector<int>> idsForTable;
        QHash<int, int> tableForId;

        while(query.next()) {
            int table = query.value(0).toInt();
            const int pair = query.value(1).toInt();
            const int player = query.value(2).toInt();
            const int id = query.value(3).toInt();

            if (pair == 2 && round > 1) {
                // If round 1 is requested, we use the draw set via the players page.
                // No table shifting happens for pair 2 in this case, so we use the respective
                // value as-is for both pair 1 and 2.
                // Otherwise, we calculate the current table for pair 2 here:
                table = table % m_tablesCount + 1;
            }

            if (! idsForTable.contains(table)) {
                idsForTable.insert(table, { 0, 0, 0, 0 });
            }

            int index = pair == 1 ? 0 : 2;
            index += player != 0 ? player - 1
                                 : idsForTable[table].at(index) != 0;

            idsForTable[table][index] = id;
            tableForId.insert(id, table);
        }

        // Assemble the assignment struct
        result.idsForTable = idsForTable;
        result.tableForId = tableForId;
    }, success);

    return { result, success };
}

TSuccess<Scores::Score> Database::scores(int round)
{
    Scores::Score score;
    bool success;

    readHelper([this, round, &score](QSqlQuery &query) {
        switch (m_tournamentSettings->tournamentMode()) {

        case Tournament::Mode::FixedPairs:
            queryPrepare(query, QStringLiteral(
                "SELECT "
                    "score.table_number, "
                    "CASE "
                    "WHEN players.disqualified IS NULL OR players.disqualified > :round THEN "
                        "players.name "
                    "ELSE "
                        "players.name || :disqualified "
                    "END, "
                    "score.score "
                "FROM score "
                "INNER JOIN players ON "
                    "players.id = score.pair_or_player "
                "INNER JOIN assignment ON "
                    "assignment.round = score.round "
                    "AND assignment.pair_or_player = score.pair_or_player "
                "WHERE score.round = :round "
                "ORDER BY "
                    "score.table_number, "
                    "assignment.pair_number, "
                    "score.booger"
            ));
            query.bindValue(QStringLiteral(":disqualified"), tr(" (disqualifiziert)"));
            query.bindValue(QStringLiteral(":round"), round);
            break;

        case Tournament::Mode::SinglePlayers:
            // It's a bit more complicated for single players ;-)
            queryPrepare(query, QStringLiteral(
                "SELECT "
                    "score_one.table_number, "
                    "CASE "
                    "WHEN player_one.disqualified IS NULL OR player_one.disqualified > :round THEN "
                        "player_one.name "
                    "ELSE "
                        "player_one.name || :disqualified "
                    "END "
                    "|| :names_separator || "
                    "CASE "
                    "WHEN player_two.disqualified IS NULL OR player_two.disqualified > :round THEN "
                        "player_two.name "
                    "ELSE "
                        "player_two.name || :disqualified "
                    "END, "
                    "score_one.score "
                "FROM "
                    "score AS score_one "
                "INNER JOIN players AS player_one ON "
                    "player_one.id = score_one.pair_or_player "
                "INNER JOIN assignment AS assignment_one ON "
                    "assignment_one.round = score_one.round "
                    "AND assignment_one.table_number = score_one.table_number "
                    "AND assignment_one.pair_or_player = score_one.pair_or_player "
                "INNER JOIN players AS player_two ON "
                    "player_two.id = score_two.pair_or_player "
                "INNER JOIN score AS score_two ON "
                    "score_two.round = score_one.round "
                    "AND score_two.table_number = score_one.table_number "
                    "AND score_two.booger = score_one.booger "
                    "AND assignment_two.pair_number = assignment_one.pair_number "
                    "AND assignment_two.player_number = 2 "
                "INNER JOIN assignment AS assignment_two ON "
                    "assignment_two.round = score_one.round "
                    "AND assignment_two.table_number = score_one.table_number "
                    "AND assignment_two.pair_or_player = score_two.pair_or_player "
                "WHERE "
                    "score_one.round = :round "
                    "AND assignment_one.player_number = 1 "
                "ORDER BY "
                    "score_one.table_number, "
                    "assignment_one.pair_number, "
                    "score_one.booger "
            ));
            query.bindValue(QStringLiteral(":disqualified"), tr(" (disqualifiziert)"));
            query.bindValue(QStringLiteral(":names_separator"), *m_namesSeparator);
            query.bindValue(QStringLiteral(":round"), round);
            break;
        }

        queryExec(query);

        const int boogersPerRound = m_tournamentSettings->boogersPerRound();
        const int boogersPerTable = boogersPerRound * 2;
        int row = 0;

        QVector<int> pairResults;
        pairResults.reserve(boogersPerRound);

        while (query.next()) {
            if (row % (boogersPerTable) == 0) {
                score.tables.append(query.value(0).toInt());
            }

            if (row % boogersPerRound == 0) {
                score.pairs.append(query.value(1).toString());
            }

            row++;

            pairResults.append(query.value(2).toInt());
            if (row % boogersPerRound == 0) {
                score.scores.append(pairResults);
                pairResults.clear();
            }
        }

    }, success);

    return { score, success };
}

TSuccess<QVector<Database::BoogerResult>> Database::singleScore(int round, int table)
{
    QVector<BoogerResult> result;
    bool success;

    readHelper([this, round, table, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral(
            "SELECT "
                "score.pair_or_player, "
                "score.score "
            "FROM score "
            "INNER JOIN assignment ON "
                "assignment.round = score.round "
                "AND assignment.pair_or_player = score.pair_or_player "
            "WHERE "
                "score.round = ? "
                "AND score.table_number = ? "
            "ORDER BY "
                "assignment.pair_number, "
                "assignment.player_number, "
                "score.booger"
        ));
        query.bindValue(0, round);
        query.bindValue(1, table);
        queryExec(query);

        while (query.next()) {
            BoogerResult boogerResult;
            boogerResult.pairOrPlayer = query.value(0).toInt();
            boogerResult.score = query.value(1).toInt();
            result.append(boogerResult);
        }
    }, success);

    return { result, success };
}

TSuccess<QVector<Database::RankData>> Database::ranking(int upToRound)
{
    QVector<RankData> result;
    bool success;

    if (upToRound == -1) {
        upToRound = roundsState().count();
    }

    readHelper([this, upToRound, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral(
            "SELECT "
                "players.name, "
                "SUM(CASE "
                        "WHEN one_score.score = :booger_score THEN 1 "
                        "ELSE 0 "
                    "END) AS won_boogers, "
                "SUM(CASE "
                        "WHEN one_score.score = :booger_score THEN 0 "
                        "ELSE one_score.score "
                    "END) AS scored_goals, "
                "SUM(CASE "
                        "WHEN other_score.score = :booger_score THEN 0 "
                        "ELSE other_score.score "
                    "END) AS opponent_goals, "
                "COUNT(one_score.score) AS played_boogers "
            "FROM "
                "score AS one_score "
            "INNER JOIN players ON "
                "players.id = one_score.pair_or_player "
            "INNER JOIN assignment AS one_assignment ON "
                "one_assignment.round = one_score.round "
                "AND one_assignment.table_number = one_score.table_number "
                "AND one_assignment.pair_or_player = one_score.pair_or_player "
            "INNER JOIN assignment AS other_assignment ON "
                "other_assignment.round = other_score.round "
                "AND other_assignment.table_number = other_score.table_number "
                "AND other_assignment.pair_or_player = other_score.pair_or_player "
            "INNER JOIN score AS other_score ON "
                "other_score.round = one_score.round "
                "AND other_score.table_number = one_score.table_number "
                "AND other_score.booger = one_score.booger "
                "AND other_assignment.player_number = one_assignment.player_number "
                "AND other_assignment.pair_number != one_assignment.pair_number "
            "WHERE "
                "one_score.round <= :up_to_round "
                "AND ("
                    "players.disqualified IS NULL "
                    "OR players.disqualified > :up_to_round"
                ") "
            "GROUP BY "
                "one_score.pair_or_player "
            "ORDER BY "
                "won_boogers DESC, "
                "scored_goals DESC, "
                "opponent_goals, "
                "players.name"
        ));
        query.bindValue(QStringLiteral(":booger_score"), m_tournamentSettings->boogerScore());
        query.bindValue(QStringLiteral(":up_to_round"), upToRound);
        queryExec(query);

        int rankingLine = 0;
        RankData lastRank;
        lastRank.rank = 0;
        lastRank.boogers = 0;
        lastRank.goals = 0;
        lastRank.opponentGoals = 0;

        while (query.next()) {
            RankData currentRank;
            currentRank.name = query.value(0).toString();
            currentRank.boogers = query.value(1).toInt();
            currentRank.goals = query.value(2).toInt();
            currentRank.opponentGoals = query.value(3).toInt();
            currentRank.playedBoogers = query.value(4).toInt();

            rankingLine++;
            if (ranksDiffer(currentRank, lastRank)) {
                currentRank.rank = rankingLine;
            } else {
                currentRank.rank = lastRank.rank;
            }

            result.append(currentRank);
            lastRank = currentRank;
        }

        if (result.count() < m_pairsOrPlayersCount) {
            // Some pair(s)/player(s) don't have a score yet,
            // so we insert empty scores to get a complete ranking list

            queryExec(query, QStringLiteral("SELECT players.name FROM players "
                                            "WHERE NOT EXISTS ( "
                                                "SELECT score.pair_or_player "
                                                "FROM score "
                                                "WHERE players.id = score.pair_or_player "
                                            ")"));

            QStringList missingNames;
            while (query.next()) {
                missingNames.append(query.value(0).toString());
            }
            m_stringTools->sort(missingNames);

            RankData missingRank;
            missingRank.boogers = 0;
            missingRank.goals = 0;
            missingRank.opponentGoals = 0;
            missingRank.playedBoogers = 0;
            if (ranksDiffer(missingRank, lastRank)) {
                missingRank.rank = lastRank.rank + 1;
            } else {
                missingRank.rank = lastRank.rank;
            }

            for (const QString &name : std::as_const(missingNames)) {
                missingRank.name = name;
                result.append(missingRank);
            }
        }
    }, success);

    return { result, success };
}

TSuccess<QVector<Markers::Marker>> Database::markers()
{
    QVector<Markers::Marker> result;
    bool success;

    readHelper([this, &result](QSqlQuery &query) {
        queryExec(query, QStringLiteral("SELECT id, name, color, sorting FROM markers "
                                        "ORDER BY sequence"));
        while (query.next()) {
            Markers::Marker marker;
            marker.id = query.value(0).toInt();
            marker.name = query.value(1).toString();
            marker.color = QColor(query.value(2).toInt());
            marker.sorting = static_cast<Markers::MarkerSorting>(query.value(3).toInt());
            result.append(marker);
        }
    }, success);

    return { result, success };
}

TSuccess<QMap<int, QVector<QString>>> Database::disqualifiedPlayers(int round)
{
    if (round == -1) {
        round = roundsState().count();
    }

    QHash<QString, int> result;
    bool success;

    readHelper([this, round, &result](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("SELECT name, disqualified FROM players "
                                           "WHERE disqualified IS NOT NULL "
                                           "AND disqualified <= :round"));
        query.bindValue(QStringLiteral(":round"), round);
        query.exec();

        while (query.next()) {
            result.insert(query.value(0).toString(), query.value(1).toInt());
        }
    }, success);

    if (! success) {
        return { { }, success };
    }

    QMap<int, QVector<QString>> sortedResult;

    // Group all names by round
    QHash<QString, int>::const_iterator resultIterator = result.constBegin();
    while (resultIterator != result.constEnd()) {
        sortedResult[resultIterator.value()].append(resultIterator.key());
        resultIterator++;
    }

    // Oder all names alphabetically for all rounds
    QMap<int, QVector<QString>>::iterator sortIterator = sortedResult.begin();
    while (sortIterator != sortedResult.end()) {
        auto list = sortIterator.value();
        m_stringTools->sort(list);
        sortedResult[sortIterator.key()] = list;
        sortIterator++;
    }

    return { sortedResult, success };
}

TSuccess<bool> Database::assignmentPresent()
{
    bool result;
    bool success;

    readHelper([this, &result](QSqlQuery &query) {
        queryExec(query, QStringLiteral("select COUNT(pair_or_player) > 0 "
                                        "FROM assignment "
                                        "WHERE round = 0;"));
        query.next();
        result = query.value(0).toBool();
    }, success);

    if (! success) {
        return { { }, success };
    }

    return { result, success };
}

// Functions writing data
// ======================

bool Database::reset()
{
    if (! writeHelper([this](QSqlQuery &query) {
        queryExec(query, QStringLiteral("DELETE FROM players"));
        queryExec(query, QStringLiteral("DELETE FROM markers"));
        queryExec(query, QStringLiteral("DELETE FROM assignment"));
        queryExec(query, QStringLiteral("DELETE FROM score"));
    })) {
        return false;
    }

    m_pairsOrPlayersCount = 0;
    m_scoresPerRound = 0;
    m_roundsState = QMap<int, bool> { { 1, false } };
    m_tournamentStarted = false;

    return true;
}

bool Database::saveSettings()
{
    if (! isWritable()
        || (! isChangeable() && ! m_tournamentSettings->tournamentOpenChanged())) {

        return true;
    }

    return writeHelper([this](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("UPDATE config SET value = ? WHERE key = ?"));
        query.bindValue(0, m_tournamentSettings->toString());
        query.bindValue(1, s_keySettings);
        queryExec(query);
        queryPrepare(query, QStringLiteral("UPDATE config SET value = ? WHERE key = ?"));
        if (m_tournamentSettings->version()
                >= m_tournamentSettings->dbTournamentSettingsVersion()) {
            // The database's settings revision is equal or older than the one used here.
            // Thus, we can simply save the settings and leave the revision as-is or update it.
            query.bindValue(0, m_tournamentSettings->version());
        } else {
            // The database's settings revision is newer than the one used here. We have unknown
            // keys which have been re-added as-is. Thus, we leave the original revision.
            query.bindValue(0, m_tournamentSettings->dbTournamentSettingsVersion());
        }
        query.bindValue(1, s_keySettingsVersion);
        queryExec(query);
    });
}

bool Database::registerPlayers(const QString &name, int markerId)
{
    return writeHelper([this, &name, markerId](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("INSERT INTO players(id, name, marker) "
                                           "VALUES(NULL, ?, ?)"));
        query.bindValue(0, name);
        query.bindValue(1, markerId != 0 ? markerId : s_sqliteNull);
        queryExec(query);
    });
}

bool Database::registerPlayers(const QStringList &names)
{
    return writeHelper([this, &names](QSqlQuery &query) {
        registerPlayers(query, names);
    });
}

void Database::registerPlayers(QSqlQuery &query, const QStringList &names)
{
    queryPrepare(query, QStringLiteral("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.addBindValue(stringListToVariantList(names));
    queryExecBatch(query);
}

bool Database::deletePlayers(const QString &name)
{
    return writeHelper([this, &name](QSqlQuery &query) {
        deletePlayers(query, name);
    });
}

void Database::deletePlayers(QSqlQuery &query, const QString &name)
{
    queryPrepare(query, QStringLiteral("DELETE FROM players WHERE name = ?"));
    query.bindValue(0, name);
    queryExec(query);
}

bool Database::renamePlayers(const QString &oldName, const QString &newName, int newMarker)
{
    return writeHelper([this, &oldName, &newName, newMarker](QSqlQuery &query) {
        renamePlayers(query, oldName, newName, newMarker);
    });
}

void Database::renamePlayers(QSqlQuery &query, const QString &oldName, const QString &newName,
                             int newMarker)
{
    queryPrepare(query, QStringLiteral("UPDATE players SET name = ? WHERE name = ?"));
    query.bindValue(0, newName);
    query.bindValue(1, oldName);
    queryExec(query);

    if (newMarker != -1) {
        updateMarker(query, newName, newMarker);
    }
}

bool Database::renamePlayers(const QStringList &oldNames, const QStringList &newNames)
{
    return writeHelper([this, &oldNames, &newNames](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("UPDATE players SET name = ? WHERE name = ?"));
        query.addBindValue(stringListToVariantList(newNames));
        query.addBindValue(stringListToVariantList(oldNames));
        queryExecBatch(query);
    });
}

bool Database::saveScore(int round, int table, const QVector<Database::BoogerResult> &score)
{
    const bool scoreSaved = writeHelper([this, round, table, &score](QSqlQuery &query) {
        switch (m_tournamentSettings->tournamentMode()) {
        case Tournament::Mode::FixedPairs:
            updateAssignment(query, round,
                             score.at(0).pairOrPlayer,
                             1, 1, table);
            updateAssignment(query, round,
                             score.at(m_tournamentSettings->boogersPerRound()).pairOrPlayer,
                             2, 1, table);
            break;
        case Tournament::Mode::SinglePlayers:
            updateAssignment(query, round,
                             score.at(0).pairOrPlayer,
                             1, 1, table);
            updateAssignment(query, round,
                             score.at(m_tournamentSettings->boogersPerRound()).pairOrPlayer,
                             1, 2, table);
            updateAssignment(query, round,
                             score.at(m_tournamentSettings->boogersPerRound() * 2).pairOrPlayer,
                             2, 1, table);
            updateAssignment(query, round,
                             score.at(m_tournamentSettings->boogersPerRound() * 3).pairOrPlayer,
                             2, 2, table);
            break;
        }

        int booger = 1;
        for (const BoogerResult &boogerResult : score) {
            queryPrepare(query, QStringLiteral(
                "INSERT INTO score(round, table_number, booger, pair_or_player, score) "
                "VALUES(?, ?, ?, ?, ?)"
            ));
            query.bindValue(0, round);
            query.bindValue(1, table);
            query.bindValue(2, booger);
            query.bindValue(3, boogerResult.pairOrPlayer);
            query.bindValue(4, boogerResult.score);
            queryExec(query);

            booger++;
            if (booger > m_tournamentSettings->boogersPerRound()) {
                booger = 1;
            }
        }
    });

    if (! scoreSaved || ! updateRoundsState()) {
        return false;
    }

    m_tournamentStarted = true;
    return true;
}

void Database::updateAssignment(QSqlQuery &query, int round, int pairOrPlayer, int pairNumber,
                                int playerNumber, int table)
{
    queryPrepare(query, QStringLiteral(
        "INSERT INTO assignment(round, pair_or_player, pair_number, player_number, table_number) "
        "VALUES(?, ?, ?, ?, ?)"
    ));
    query.bindValue(0, round);
    query.bindValue(1, pairOrPlayer);
    query.bindValue(2, pairNumber);
    query.bindValue(3, playerNumber);
    query.bindValue(4, table);
    queryExec(query);
}

bool Database::deleteScore(int round, int table)
{
    const bool scoreDeleted = writeHelper([this, round, table](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("DELETE FROM score "
                                           "WHERE round = ? "
                                           "AND table_number = ?"));
        query.bindValue(0, round);
        query.bindValue(1, table);
        queryExec(query);

        queryPrepare(query, QStringLiteral("DELETE FROM assignment "
                                           "WHERE round = ? "
                                           "AND table_number = ?"));
        query.bindValue(0, round);
        query.bindValue(1, table);
        queryExec(query);
    });

    if (! scoreDeleted || ! checkIfTournamentStarted() || ! updateRoundsState()) {
        return false;
    }

    return true;
}

bool Database::addMarkers(const QVector<Markers::Marker> &markers)
{
    return writeHelper([this, &markers](QSqlQuery &query) {
        for (const Markers::Marker &marker : markers) {
            addMarker(query,
                    marker.name.isEmpty() ? QString() : marker.name,
                    marker.color,
                    marker.sorting,
                    marker.id);
        }
    });
}

bool Database::addMarker(const QString &name, const QColor &color, Markers::MarkerSorting sorting)
{
    return writeHelper([this, &name, &color, sorting](QSqlQuery &query) {
        addMarker(query, name, color, sorting);
    });
}

void Database::addMarker(QSqlQuery &query, const QString &name, const QColor &color,
                         Markers::MarkerSorting sorting, int id)
{
    queryPrepare(query, QStringLiteral("INSERT INTO markers(id, name, color, sorting, sequence) "
                                       "VALUES(?, ?, ?, ?, "
                                              "(SELECT COUNT(sequence) FROM markers))"));
    query.bindValue(0, id != -1 ? id : s_sqliteNull);
    query.bindValue(1, name);
    query.bindValue(2, color.rgb());
    query.bindValue(3, sorting);
    queryExec(query);
}

bool Database::deleteMarker(int id, int sequence)
{
    return writeHelper([this, id, sequence](QSqlQuery &query) {
        // Delete the marker
        queryPrepare(query, QStringLiteral("DELETE FROM markers WHERE id = ?"));
        query.bindValue(0, id);
        queryExec(query);

        // Close gaps in the sequence counter
        queryPrepare(query, QStringLiteral("UPDATE markers SET sequence = sequence - 1 "
                                           "WHERE sequence > ?"));
        query.bindValue(0, sequence);
        queryExec(query);

        // Remove all respective marker links from the pairs/players list
        queryPrepare(query, QStringLiteral("UPDATE players SET marker = NULL WHERE marker = ?"));
        query.bindValue(0, id);
        queryExec(query);
    });
}

bool Database::editMarker(int id, const QString &name, const QColor &color,
                          Markers::MarkerSorting sorting)
{
    return writeHelper([this, id, &name, &color, sorting](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("UPDATE markers SET name = ?, color = ?, sorting = ? "
                                           "WHERE id = ?"));
        query.bindValue(0, name);
        query.bindValue(1, color.rgb());
        query.bindValue(2, sorting);
        query.bindValue(3, id);
        queryExec(query);
    });
}

bool Database::moveMarker(int id, int sequence, int direction)
{
    return writeHelper([this, id, sequence, direction](QSqlQuery &query) {
        if ((sequence == 0 && direction == 1) || (sequence == 1 && direction == -1)) {
            // Be sure that the new first marker will create a new block
            queryPrepare(query, QStringLiteral("UPDATE markers SET sorting = ? "
                                               "WHERE sequence = 1"));
            query.bindValue(0, Markers::MarkerSorting::NewBlock);
            queryExec(query);
        }

        queryPrepare(query, QStringLiteral("UPDATE markers SET sequence = sequence - ? "
                                           "WHERE sequence = ?"));
        query.bindValue(0, direction);
        query.bindValue(1, sequence + direction);
        queryExec(query);

        queryPrepare(query, QStringLiteral("UPDATE markers SET sequence = ? WHERE id = ?"));
        query.bindValue(0, sequence + direction);
        query.bindValue(1, id);
        queryExec(query);
    });
}

bool Database::updateMarker(const QString &name, int markerId)
{
    return writeHelper([this, &name, markerId](QSqlQuery &query) {
        updateMarker(query, name, markerId);

        if (markerId == m_tournamentSettings->marker(Markers::Absent)) {
            updateDraw(query, name, Draw::EmptySeat, -1);
        }
    });
}

void Database::updateMarker(QSqlQuery &query, const QString &name, int markerId)
{
    queryPrepare(query, QStringLiteral("UPDATE players SET marker = ? WHERE name = ?"));
    query.bindValue(0, markerId != 0 ? markerId : s_sqliteNull);
    query.bindValue(1, name);
    queryExec(query);
}

bool Database::listUpdateMarker(int markerId)
{
    return writeHelper([this, markerId](QSqlQuery &query) {
        listUpdateMarker(query, markerId);

        if (markerId == m_tournamentSettings->marker(Markers::Absent)) {
            deleteAllDraws(query);
        }
    });
}

void Database::listUpdateMarker(QSqlQuery &query, int markerId)
{
    queryPrepare(query, QStringLiteral("UPDATE players SET marker = ?"));
    query.bindValue(0, markerId != 0 ? markerId : s_sqliteNull);
    queryExec(query);
}

bool Database::listRemoveMarker(int markerId)
{
    return writeHelper([this, markerId](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("UPDATE players SET marker = NULL where marker = ?"));
        query.bindValue(0, markerId);
        queryExec(query);
    });
}

bool Database::deleteMarkedPlayers(int markerId)
{
    return writeHelper([this, markerId](QSqlQuery &query) {
        queryPrepare(query, QStringLiteral("DELETE FROM players where marker = ?"));
        query.bindValue(0, markerId);
        queryExec(query);
    });
}

bool Database::deleteUnmarkedPlayers()
{
    return writeHelper([this](QSqlQuery &query) {
        queryExec(query, QStringLiteral("DELETE FROM players where marker IS NULL"));
    });
}

bool Database::deleteAllPlayers()
{
    return writeHelper([this](QSqlQuery &query) {
        queryExec(query, QStringLiteral("DELETE FROM players"));
    });
}

bool Database::adoptAllData(const QHash<QString, PlayersData> &playersData,
                            const QVector<Markers::Marker> &markers)
{
    return writeHelper([this, &playersData, &markers](QSqlQuery &query) {
        // Adopt the players list
        queryExec(query, QStringLiteral("DELETE FROM players"));
        const QStringList names(playersData.keys());
        registerPlayers(query, names);

        // Adopt the markers and the draws
        adoptPlayersMetadata(query, names, playersData, markers);
    });
}

bool Database::adoptPlayersMetadata(const QHash<QString, PlayersData> &playersData,
                                    const QVector<Markers::Marker> &markers)
{
    return writeHelper([this, &playersData, &markers](QSqlQuery &query) {
        const QStringList names(playersData.keys());
        adoptPlayersMetadata(query, names, playersData, markers);
    });
}

void Database::adoptPlayersMetadata(QSqlQuery &query,
                                    const QStringList &names,
                                    const QHash<QString, PlayersData> &playersData,
                                    const QVector<Markers::Marker> &markers)
{
    listUpdateMarker(query, 0);
    deleteAllDraws(query);

    // Adpot all markers
    queryExec(query, QStringLiteral("DELETE FROM markers"));
    for (const Markers::Marker &marker : markers) {
        addMarker(query, marker.name,  marker.color, marker.sorting, marker.id);
    }

    for (const QString &name : names) {
        // Adopt the marker if set
        const int markerId = playersData.value(name).marker;
        if (markerId != 0) {
            updateMarker(query, name, markerId);
        }

        // Adopt the draw if set
        const int assignedTable = playersData.value(name).assignedTable;
        if (assignedTable != 0) {
            queryPrepare(query, QStringLiteral(
                "INSERT INTO assignment(round, pair_or_player, table_number, pair_number) "
                "VALUES(0, (SELECT id FROM players WHERE name = ?), ?, ?)"));
            query.bindValue(0, name);
            query.bindValue(1, assignedTable);
            query.bindValue(2, playersData.value(name).assignedPairNumber);
            queryExec(query);
        }

        // Adopt the disqualification if set
        const int disqualified = playersData.value(name).disqualified;
        if (disqualified != 0) {
            setDisqualified(query, name, disqualified);
        }
    }
}

bool Database::assemblePair(const QString &player1, const QString &player2,
                            const QString &separator, int assignedMarker)
{
    return writeHelper([this, &player1, &player2, &separator, assignedMarker](QSqlQuery &query) {
        const QString pairName = player1 + separator + player2;
        renamePlayers(query, player1, pairName, assignedMarker);
        deletePlayers(query, player2);
    });
}

bool Database::updateDraw(const QString &name, const Draw::Seat &seat, int marker)
{
    return writeHelper([this, &name, &seat, marker](QSqlQuery &query) {
        updateDraw(query, name, seat, marker);
    });
}

void Database::updateDraw(QSqlQuery &query, const QString &name, const Draw::Seat &seat,
                                   int marker)
{
    // Cache the pair's/player's id (sadly, we have to do it this way, because if we're on a
    // network, the ids of the registrations can differ. Thus, we have to query for the name)
    queryPrepare(query, QStringLiteral("SELECT id from players WHERE name = ?"));
    query.bindValue(0, name);
    queryExec(query);
    query.next();
    const int id = query.value(0).toInt();

    // Remove a possibly existing round 0 assignment
    queryPrepare(query, QStringLiteral("DELETE FROM assignment "
                                       "WHERE round = 0 AND pair_or_player = ?"));
    query.bindValue(0, id);
    queryExec(query);

    // Then, if a table was given, insert the assignment
    if (! seat.isEmpty()) {
        queryPrepare(query, QStringLiteral(
            "INSERT INTO assignment(round, pair_or_player, table_number, pair_number) "
            "VALUES(0, ?, ?, ?)"));
        query.bindValue(0, id);
        query.bindValue(1, seat.table);
        query.bindValue(2, seat.pair);
        queryExec(query);

        if (marker != -1) {
            updateMarker(query, name, marker);
        }
    }
}

bool Database::setListDraw(const QVector<QString> &names, const QVector<Draw::Seat> &seats)
{
    return writeHelper([this, &names, &seats](QSqlQuery &query) {
        // We have to get all IDs first. Operations on players have to queried via their name,
        // because the ID can differ between machines if we're on a network.
        QHash<QString, int> idMap;
        queryPrepare(query, QStringLiteral("SELECT name, id FROM players;"));
        queryExec(query);
        while (query.next()) {
            idMap.insert(query.value(0).toString(), query.value(1).toInt());
        }

        // Remove all draws
        deleteAllDraws(query);

        // Create a QVariantList of all values we want to insert
        QVariantList ids;
        QVariantList tables;
        QVariantList pairs;
        for (int i = 0; i < names.count(); i++) {
            ids.append(idMap.value(names.at(i)));
            tables.append(seats.at(i).table);
            pairs.append(seats.at(i).pair);
        }

        // Insert the data
        queryPrepare(query, QStringLiteral(
            "INSERT INTO assignment(round, pair_or_player, table_number, pair_number) "
            "VALUES(0, ?, ?, ?)"));
        query.addBindValue(ids);
        query.addBindValue(tables);
        query.addBindValue(pairs);
        queryExecBatch(query);
    });
}

bool Database::deleteAllDraws()
{
    return writeHelper([this](QSqlQuery &query) {
        deleteAllDraws(query);
    });
}

void Database::deleteAllDraws(QSqlQuery &query)
{
    queryExec(query, QStringLiteral("DELETE FROM assignment WHERE round = 0"));
}

bool Database::setDisqualified(const QString &name, int round)
{
    return writeHelper([this, &name, round](QSqlQuery &query) {
        setDisqualified(query, name, round);
    });
}

void Database::setDisqualified(QSqlQuery &query, const QString &name, int round)
{
    queryPrepare(query, QStringLiteral("UPDATE players "
                                        "SET disqualified = :disqualified_value "
                                        "WHERE name = :name"));
    query.bindValue(QStringLiteral(":disqualified_value"), round != 0 ? round : s_sqliteNull);
    query.bindValue(QStringLiteral(":name"), name);
    queryExec(query);
}

// =======================
// Functions not using SQL
// =======================

QVariantList Database::stringListToVariantList(const QStringList &stringList) const
{
    QVariantList variantList;
    for (const QString &text : stringList) {
        variantList.append(QVariant(text));
    }
    return variantList;
}

const QString &Database::errorMessage() const
{
    return m_errorMessage;
}

bool Database::isOpen() const
{
    return m_db.isOpen();
}

bool Database::isWritable() const
{
    return m_isWritable;
}

bool Database::isChangeable() const
{
    return m_isWritable && m_tournamentSettings->tournamentOpen();
}

const QString &Database::dbFile() const
{
    return m_dbFileName;
}

const QString &Database::dbPath() const
{
    return m_dbFilePath;
}

int Database::dbVersion() const
{
    return s_dbVersion;
}

void Database::setNetworkTournamentStarted(bool state)
{
    m_networkTournamentStarted = state;
}

bool Database::networkTournamentStarted() const
{
    return m_networkTournamentStarted;
}

bool Database::tournamentStarted() const
{
    return m_tournamentStarted || m_networkTournamentStarted;
}

bool Database::localTournamentStarted() const
{
    return m_tournamentStarted;
}

int Database::necessaryDivisor() const
{
    return m_tournamentSettings->isPairMode() ? 2 : 4;
}

bool Database::playersDividable() const
{
    return playersDividable(m_pairsOrPlayersCount);
}

bool Database::playersDividable(int count) const
{
    return count > 0 && count % necessaryDivisor() == 0;
}

void Database::setPairsOrPlayersCount(int count)
{
    m_pairsOrPlayersCount = count;
    m_scoresPerRound = m_pairsOrPlayersCount * m_tournamentSettings->boogersPerRound();
    if (playersDividable()) {
        m_tablesCount = m_pairsOrPlayersCount / necessaryDivisor();
    }
}

int Database::pairsOrPlayersCount() const
{
    return m_pairsOrPlayersCount;
}

bool Database::ranksDiffer(const RankData &currentRank, const RankData &lastRank)
{
    if (m_tournamentSettings->includeOpponentGoals()) {
        return currentRank.boogers != lastRank.boogers
               || currentRank.goals != lastRank.goals
               || currentRank.opponentGoals != lastRank.opponentGoals;
    } else {
        return currentRank.boogers != lastRank.boogers
               || currentRank.goals != lastRank.goals;
    }
}

const QMap<int, bool> &Database::roundsState() const
{
    return m_roundsState;
}

bool Database::allRoundsFinished() const
{
    if (! tournamentStarted()) {
        return true;
    }

    const auto roundsStateValues = m_roundsState.values();
    for (const auto finished : roundsStateValues) {
        if (! finished) {
            return false;
        }
    }

    return true;
}

int Database::allRounds() const
{
    return m_allRounds;
}

Database::GoalsImportant Database::goalsImportance(const QVector<Database::RankData> &ranking) const
{
    QVector<bool> goalsImportance;
    QVector<bool> opponentGoalsImportance;

    for (int index = 0; index < ranking.count(); index++) {
        // Get the boogers and goals of the previous rank

        int prevRankBoogers = -1;
        int prevRankGoals = -1;
        int prevRankIndex = index - 1;

        while (prevRankIndex > 0 && ranking.at(prevRankIndex).rank == ranking.at(index).rank) {
            prevRankIndex--;
        }

        if (prevRankIndex >= 0 && ranking.at(prevRankIndex).rank != ranking.at(index).rank) {
            prevRankBoogers = ranking.at(prevRankIndex).boogers;
            prevRankGoals = ranking.at(prevRankIndex).goals;
        }

        // Get the boogers and goals of the next rank

        int nextRankBoogers = -1;
        int nextRankGoals = -1;
        int nextRankIndex = index + 1;

        while (nextRankIndex < ranking.count()
               && ranking.at(nextRankIndex).rank == ranking.at(index).rank) {

            nextRankIndex++;
        }

        if (nextRankIndex < ranking.count()
            && ranking.at(nextRankIndex).rank != ranking.at(index).rank) {

            nextRankBoogers = ranking.at(nextRankIndex).boogers;
            nextRankGoals = ranking.at(nextRankIndex).goals;
        }

        // Check what's important

        bool goalsImportant = true;
        bool opponentGoalsImportant = true;

        if (ranking.at(index).boogers != prevRankBoogers
            && ranking.at(index).boogers != nextRankBoogers) {

            goalsImportant = false;
            opponentGoalsImportant = false;

        } else if (ranking.at(index).goals != prevRankGoals
                   && ranking.at(index).goals != nextRankGoals) {

            opponentGoalsImportant = false;
        }

        goalsImportance.append(goalsImportant);
        opponentGoalsImportance.append(opponentGoalsImportant);
    }

    return { goalsImportance, opponentGoalsImportance };
}
